﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.EGO {
  internal static class EgoUtilities {
    public static int ArgMax<T>(this IEnumerable<T> values, Func<T, double> func) {
      var max = double.MinValue;
      var maxIdx = 0;
      var idx = 0;
      foreach (var v in values) {
        var d = func.Invoke(v);
        if (d > max) {
          max = d;
          maxIdx = idx;
        }
        idx++;
      }
      return maxIdx;
    }
    public static int ArgMin<T>(this IEnumerable<T> values, Func<T, double> func) {
      return ArgMax(values, x => -func.Invoke(x));
    }

    public static ResultCollection SyncRunSubAlgorithm(IAlgorithm alg, int random) {

      if (alg.Parameters.ContainsKey("SetSeedRandomly") && alg.Parameters.ContainsKey("Seed")) {
        var setSeed = alg.Parameters["SetSeedRandomly"].ActualValue as BoolValue;
        var seed = alg.Parameters["Seed"].ActualValue as IntValue;
        if (seed == null || setSeed == null) throw new ArgumentException("wrong SeedParametertypes");
        setSeed.Value = false;
        seed.Value = random;

      }


      EventWaitHandle trigger = new AutoResetEvent(false);
      Exception ex = null;
      EventHandler<EventArgs<Exception>> exhandler = (sender, e) => ex = e.Value;
      EventHandler stoppedHandler = (sender, e) => trigger.Set();
      alg.ExceptionOccurred += exhandler;
      alg.Stopped += stoppedHandler;
      alg.Prepare();
      alg.Start();
      trigger.WaitOne();
      alg.ExceptionOccurred -= exhandler;
      alg.Stopped -= stoppedHandler;
      if (ex != null) throw ex;
      return alg.Results;
    }
    public static RealVector[] GetUniformRandomDesign(int points, int dim, DoubleMatrix bounds, IRandom random) {
      var res = new RealVector[points];
      for (var i = 0; i < points; i++) {
        var r = new RealVector(dim);
        res[i] = r;
        for (var j = 0; j < dim; j++) {
          var b = j % bounds.Rows;
          r[j] = UniformRandom(bounds[b, 0], bounds[b, 1], random);
        }
      }
      return res;
    }
    public static double UniformRandom(double min, double max, IRandom rand) {
      return rand.NextDouble() * (max - min) + min;
    }

    public static double GetEstimation(this IRegressionModel model, RealVector r) {
      var dataset = GetDataSet(new[] { new Tuple<RealVector, double>(r, 0.0) }, false);
      return model.GetEstimatedValues(dataset, new[] { 0 }).First();
    }
    public static double GetVariance(this IConfidenceRegressionModel model, RealVector r) {
      var dataset = GetDataSet(new[] { new Tuple<RealVector, double>(r, 0.0) }, false);
      return model.GetEstimatedVariances(dataset, new[] { 0 }).First();
    }


    public static double GetDoubleValue(this IDataset dataset, int i, int j) {
      return dataset.GetDoubleValue("input" + j, i);
    }
    public static Dataset GetDataSet(IReadOnlyList<Tuple<RealVector, double>> samples, bool removeDuplicates) {
      if (removeDuplicates)
        samples = RemoveDuplicates(samples); //TODO duplicates require heteroskedasticity in Models


      var dimensions = samples[0].Item1.Length + 1;
      var data = new double[samples.Count, dimensions];
      var names = new string[dimensions - 1];
      for (var i = 0; i < names.Length; i++) names[i] = "input" + i;

      for (var j = 0; j < samples.Count; j++) {
        for (var i = 0; i < names.Length; i++) data[j, i] = samples[j].Item1[i];
        data[j, dimensions - 1] = samples[j].Item2;

      }


      return new Dataset(names.Concat(new[] { "output" }).ToArray(), data);
    }

    private static IReadOnlyList<Tuple<RealVector, double>> RemoveDuplicates(IReadOnlyList<Tuple<RealVector, double>> samples) {
      var res = new List<Tuple<RealVector, double, int>>();

      foreach (var sample in samples) {
        if (res.Count == 0) {
          res.Add(new Tuple<RealVector, double, int>(sample.Item1, sample.Item2, 1));
          continue;
        }

        var index = res.ArgMin(x => Euclidian(sample.Item1, x.Item1));
        var d = Euclidian(res[index].Item1, sample.Item1);
        if (d > 0.0001)
          res.Add(new Tuple<RealVector, double, int>(sample.Item1, sample.Item2, 1));
        else {
          var t = res[index];
          res.RemoveAt(index);
          res.Add(new Tuple<RealVector, double, int>(t.Item1, t.Item2 + sample.Item2, t.Item3 + 1));
        }
      }
      return res.Select(x => new Tuple<RealVector, double>(x.Item1, x.Item2 / x.Item3)).ToArray();
    }

    private static double Euclidian(IEnumerable<double> a, IEnumerable<double> b) {
      return Math.Sqrt(a.Zip(b, (d, d1) => d - d1).Sum(d => d * d));
    }

    public static DoubleMatrix GetBoundingBox(IEnumerable<RealVector> vectors) {
      DoubleMatrix res = null;
      foreach (var vector in vectors)
        if (res == null) {
          res = new DoubleMatrix(vector.Length, 2);
          for (var i = 0; i < vector.Length; i++)
            res[i, 0] = res[i, 1] = vector[i];
        } else
          for (var i = 0; i < vector.Length; i++) {
            res[i, 0] = Math.Min(vector[i], res[i, 0]);
            res[i, 1] = Math.Max(vector[i], res[i, 1]);
          }
      return res;
    }


  }
}
